use dashmap::DashMap;

/// A macro to simplify using a cache for API responses.
///
/// This macro checks if a value for a given route and ID exists in the cache.
/// If the value is found, it deserializes it into JSON and immediately returns it.
///
/// # Parameters
/// * `$route` - A literal string representing the route.
/// * `$id` - A variable or literal string representing the unique identifier for the cached value.
/// * `$cache` - The cache instance to query.
///
/// # Example
///
/// ```ignore
/// use based::use_api_cache;
///
/// use_api_cache!("/user", user_id, cache);
/// ```
#[macro_export]
macro_rules! use_api_cache {
    ($route:literal, $id:ident, $cache:ident) => {
        if let Some(ret) = $cache.get_only($route, $id).await {
            return Ok(serde_json::from_str(&ret).unwrap());
        }
    };
    ($route:literal, $id:literal, $cache:ident) => {
        if let Some(ret) = $cache.get_only($route, $id).await {
            return Ok(serde_json::from_str(&ret).unwrap());
        }
    };
}

/// A structure for managing cached API responses in memory.
///
/// The cache uses a nested `HashMap` structure to store values. Each route
/// maps to another `HashMap` where keys are IDs and values are cached results.
/// The cache supports asynchronous reads and writes using `RwLock`.
pub struct RouteCache {
    // Route cache with outer HashMap storing routes, inner HashMap storing IDs
    inner: DashMap<String, DashMap<String, Option<String>>>,
}

impl RouteCache {
    /// Creates a new `RouteCache` instance.
    ///
    /// # Returns
    /// A new `RouteCache` instance with an empty cache.
    pub fn new() -> Self {
        Self {
            inner: DashMap::new(),
        }
    }

    /// Retrieves a cached value or generates and caches it if not found.
    ///
    /// This method first checks if the value exists in the cache. If not, it uses the provided
    /// generator function to compute the value, caches it, and then returns it.
    ///
    /// # Arguments
    /// * `route` - The route associated with the cached value.
    /// * `id` - The unique identifier for the value.
    /// * `generator` - A function that generates the value if it's not in the cache.
    ///
    /// # Returns
    /// A `String` containing the cached or generated value.
    pub async fn get<F, Fut>(&self, route: &str, id: &str, generator: F) -> String
    where
        F: FnOnce() -> Fut,
        Fut: std::future::Future<Output = String>,
    {
        if let Some(inner_map) = self.inner.get(route) {
            if let Some(cached_value) = inner_map.get(id) {
                log::info!("Using cached value for {route} / {id}");
                return cached_value.clone().unwrap();
            }
        }

        log::info!("Computing value for {route} / {id}");
        let computed = generator().await;

        self.inner
            .entry(route.to_string())
            .or_insert_with(DashMap::new)
            .insert(id.to_string(), Some(computed.clone()));

        computed
    }

    /// Retrieves a cached value if it exists, without generating a new one.
    ///
    /// # Arguments
    /// * `route` - The route associated with the cached value.
    /// * `id` - The unique identifier for the value.
    ///
    /// # Returns
    /// An `Option<String>` containing the cached value, or `None` if not found.
    pub async fn get_only(&self, route: &str, id: &str) -> Option<String> {
        if let Some(inner_map) = self.inner.get(route) {
            if let Some(cached_value) = inner_map.get(id) {
                log::info!("Using cached value for {route} / {id}");
                return cached_value.clone();
            }
        }

        None
    }

    /// Retrieves a cached value or generates it conditionally if not found.
    ///
    /// This method works similarly to `get`, but the generator can return an
    /// `Option<String>` to handle cases where the value might not always be computable.
    ///
    /// # Arguments
    /// * `route` - The route associated with the cached value.
    /// * `id` - The unique identifier for the value.
    /// * `generator` - A function that generates the value if it's not in the cache.
    ///
    /// # Returns
    /// An `Option<String>` containing the cached or generated value.
    pub async fn get_option<F, Fut>(&self, route: &str, id: &str, generator: F) -> Option<String>
    where
        F: FnOnce() -> Fut,
        Fut: std::future::Future<Output = Option<String>>,
    {
        if let Some(inner_map) = self.inner.get(route) {
            if let Some(cached_value) = inner_map.get(id) {
                log::info!("Using cached value for {route} / {id}");
                return cached_value.clone();
            }
        }

        log::info!("Computing value for {route} / {id}");
        let computed = generator().await;

        self.inner
            .entry(route.to_string())
            .or_insert_with(DashMap::new)
            .insert(id.to_string(), computed.clone());

        computed
    }

    /// Inserts a value into the cache.
    ///
    /// # Arguments
    /// * `route` - The route associated with the value.
    /// * `id` - The unique identifier for the value.
    /// * `value` - The value to insert into the cache.
    pub async fn insert(&self, route: &str, id: &str, value: String) {
        log::info!("Inserting value for {route} / {id}");

        self.inner
            .entry(route.to_string())
            .or_insert_with(DashMap::new)
            .insert(id.to_string(), Some(value));
    }

    /// Invalidates a cached value, removing it from the cache.
    ///
    /// # Arguments
    /// * `route` - The route associated with the value to remove.
    /// * `id` - The unique identifier for the value to remove.
    pub async fn invalidate(&self, route: &str, id: &str) {
        if let Some(inner_map) = self.inner.get_mut(route) {
            inner_map.remove(id);

            // If the inner map is empty, remove the route entry as well.
            if inner_map.is_empty() {
                self.inner.remove(route);
            }
        }
    }
}
