use maud::{Render, html};

use crate::ui::UIWidget;

#[allow(non_snake_case)]
pub fn Avatar<T: Into<String>>(image: T, name: T) -> AvatarWidget {
    AvatarWidget {
        users: vec![(image.into(), name.into())],
        ring: false,
        use_initials: false,
        indicator: None,
    }
}

#[allow(non_snake_case)]
pub fn AvatarStack(users: Vec<(String, String)>) -> AvatarWidget {
    AvatarWidget {
        users,
        ring: false,
        use_initials: false,
        indicator: None,
    }
}

pub struct AvatarWidget {
    users: Vec<(String, String)>,
    ring: bool,
    use_initials: bool,
    indicator: Option<bool>,
}

impl AvatarWidget {
    pub fn with_ring(mut self) -> Self {
        self.ring = true;
        self
    }

    pub fn online(mut self, value: bool) -> Self {
        self.indicator = Some(value);
        self
    }

    pub fn use_initials(mut self) -> Self {
        self.use_initials = true;
        self
    }

    pub fn build_avatar(&self, image: &str, name: &str) -> maud::Markup {
        let mut img_class = "w-10 h-10 rounded-full".to_string();

        if self.ring {
            img_class.push_str(" ring-2 ring-gray-300 dark:ring-gray-500");
        }

        if self.indicator.is_some() {
            let online = self.indicator.unwrap();

            let indicator_class = match online {
                true => {
                    "bottom-0 left-7 absolute w-3.5 h-3.5 bg-green-400 border-2 border-white dark:border-gray-800 rounded-full"
                }
                false => {
                    "bottom-0 left-7 absolute w-3.5 h-3.5 bg-red-400 border-2 border-white dark:border-gray-800 rounded-full"
                }
            };

            return html! {
                @if image.is_empty() {
                    @if self.use_initials {
                            div class="relative" title=(name) {
                            div class=(format!("relative inline-flex items-center justify-center w-10 h-10 overflow-hidden bg-gray-100 rounded-full dark:bg-gray-600 {img_class}")) {
                                span class="font-medium text-gray-600 dark:text-gray-300" { (initials(name)) };
                            };
                            span class=(indicator_class) {};
                        }
                    } @else {
                        div class="relative" title=(name) {
                            div class=(format!("relative w-10 h-10 overflow-hidden bg-gray-100 rounded-full dark:bg-gray-600 {img_class}")){
                                svg class="absolute w-12 h-12 text-gray-400 -left-1" fill="currentColor" viewBox="0 0 20 20" xmlns="http://www.w3.org/2000/svg" {
                                    path fill-rule="evenodd" d="M10 9a3 3 0 100-6 3 3 0 000 6zm-7 9a7 7 0 1114 0H3z" clip-rule="evenodd" {};
                                };
                            };
                            span class=(indicator_class) {};
                        };
                    };
                } @else {
                div class="relative" title=(name) {
                    img class=(img_class) src=(image) alt=(name) {};
                    span class=(indicator_class) {};
                };
            }
            };
        } else {
            return html! {
                @if image.is_empty() {
                    @if self.use_initials {
                            div class="relative" title=(name) {
                            div class=(format!("relative inline-flex items-center justify-center w-10 h-10 overflow-hidden bg-gray-100 rounded-full dark:bg-gray-600 {img_class}")) {
                                span class="font-medium text-gray-600 dark:text-gray-300" { (initials(name)) };
                            };
                        }
                    } @else {
                        div class="relative" title=(name) {
                            div class=(format!("relative w-10 h-10 overflow-hidden bg-gray-100 rounded-full dark:bg-gray-600 {img_class}")){
                                svg class="absolute w-12 h-12 text-gray-400 -left-1" fill="currentColor" viewBox="0 0 20 20" xmlns="http://www.w3.org/2000/svg" {
                                    path fill-rule="evenodd" d="M10 9a3 3 0 100-6 3 3 0 000 6zm-7 9a7 7 0 1114 0H3z" clip-rule="evenodd" {};
                                };
                            };
                        };
                    };
                } @else {
                div class="relative" title=(name) {
                    img class=(img_class) src=(image) alt=(name) {};
                };
            }
            };
        }
    }
}

fn initials(txt: &str) -> String {
    let elements = txt.split_whitespace();
    let mut initials = Vec::new();

    for e in elements {
        initials.push(e.chars().next().unwrap().to_string());
    }

    initials.join("")
}

impl Render for AvatarWidget {
    fn render(&self) -> maud::Markup {
        self.render_with_class("")
    }
}

impl UIWidget for AvatarWidget {
    fn can_inherit(&self) -> bool {
        false
    }

    fn base_class(&self) -> Vec<String> {
        Vec::new()
    }

    fn extended_class(&self) -> Vec<String> {
        Vec::new()
    }

    fn render_with_class(&self, _: &str) -> maud::Markup {
        if self.users.len() == 1 {
            let (image, name) = &self.users[0];
            self.build_avatar(image, name);
        }

        html! {
            div class="flex -space-x-4 rtl:space-x-reverse select-none" {
                @for (image, name) in self.users.iter().take(4) {
                    (self.build_avatar(&image, &name))
                }

                @if self.users.len() > 4 {
                    p class="flex items-center justify-center w-11 h-11 text-xs font-medium text-white bg-gray-700 border-2 border-white rounded-full hover:bg-gray-600 dark:border-gray-800 z-10 -translate-y-[2px]" {
                        (format!("+{}", self.users.len()-4))
                    };
                }
            }
        }
    }
}
