use maud::{Render, html};

use crate::ui::UIWidget;

pub struct TimelineElement {
    icon: Option<Box<dyn UIWidget>>,
    time_heading: bool,
    title: Box<dyn UIWidget>,
    time: Box<dyn UIWidget>,
    body: Box<dyn UIWidget>,
}

impl TimelineElement {
    pub fn new<T: UIWidget + 'static, B: UIWidget + 'static, X: UIWidget + 'static>(
        time: X,
        title: T,
        body: B,
    ) -> Self {
        TimelineElement {
            icon: None,
            time_heading: false,
            title: Box::new(title),
            time: Box::new(time),
            body: Box::new(body),
        }
    }

    pub fn time_heading(mut self) -> Self {
        self.time_heading = true;
        self
    }

    pub fn icon<T: UIWidget + 'static>(mut self, icon: T) -> Self {
        self.icon = Some(Box::new(icon));
        self
    }
}

impl Render for TimelineElement {
    fn render(&self) -> maud::Markup {
        html! {
            li class="mb-10 ms-6" {
                @if let Some(icon) = &self.icon {
                span class="absolute flex items-center justify-center w-6 h-6 bg-blue-100 rounded-full -start-3 dark:ring-gray-900 dark:bg-blue-900" {
                    (icon)
                };
            } @else {
                div class="absolute w-3 h-3 bg-gray-200 rounded-full mt-1.5 -start-1.5 border border-white dark:border-gray-900 dark:bg-gray-700" {};
            }

            @if self.time_heading {
                time class="mb-1 text-sm font-normal leading-none text-gray-400 dark:text-gray-500" { (self.time) };
                h3 class="text-lg font-semibold text-gray-900 dark:text-white" { (self.title) };
            } @else {
                h3 class="mb-1 text-lg font-semibold text-gray-900 dark:text-white" { (self.title) };
                time class="block mb-2 text-sm font-normal leading-none text-gray-400 dark:text-gray-500" { (self.time) };
            }

                (self.body)
                };
        }
    }
}

pub struct TimelineWidget {
    vanish: bool,
    elements: Vec<TimelineElement>,
    after: Option<Box<dyn UIWidget>>,
}

impl TimelineWidget {
    pub fn add_element(mut self, element: TimelineElement) -> Self {
        self.elements.push(element);
        self
    }

    pub fn add_after<T: UIWidget + 'static>(mut self, after: T) -> Self {
        self.after = Some(Box::new(after));
        self
    }

    pub fn vanish(mut self) -> Self {
        self.vanish = true;
        self
    }
}

impl Render for TimelineWidget {
    fn render(&self) -> maud::Markup {
        self.render_with_class("")
    }
}

impl UIWidget for TimelineWidget {
    fn can_inherit(&self) -> bool {
        false
    }

    fn base_class(&self) -> Vec<String> {
        Vec::new()
    }

    fn extended_class(&self) -> Vec<String> {
        Vec::new()
    }

    fn render_with_class(&self, _: &str) -> maud::Markup {
        if self.vanish {
            return html! {
                @for element in &self.elements {
                    (element);
                }

                @if let Some(after) = &self.after {
                    (after)
                }
            };
        }

        html! {
            ol class="ml-6 mt-4 relative border-s border-gray-200 dark:border-gray-700 w-fit" {
                @for element in &self.elements {
                    (element);
                }

                @if let Some(after) = &self.after {
                    (after)
                }
            };
        }
    }
}

#[allow(non_snake_case)]
pub fn Timeline() -> TimelineWidget {
    TimelineWidget {
        vanish: false,
        elements: Vec::new(),
        after: None,
    }
}

pub struct ActivityLogElement {
    icon: Box<dyn UIWidget>,
    title: Box<dyn UIWidget>,
    time: Box<dyn UIWidget>,
    body: Option<Box<dyn UIWidget>>,
}

impl ActivityLogElement {
    pub fn new<T: UIWidget + 'static, X: UIWidget + 'static, I: UIWidget + 'static>(
        icon: I,
        time: X,
        title: T,
    ) -> Self {
        ActivityLogElement {
            icon: Box::new(icon),
            title: Box::new(title),
            time: Box::new(time),
            body: None,
        }
    }

    pub fn body<T: UIWidget + 'static>(mut self, body: T) -> Self {
        self.body = Some(Box::new(body));
        self
    }
}

impl Render for ActivityLogElement {
    fn render(&self) -> maud::Markup {
        html! {
            li class="mb-10 ms-6" {
                span class="absolute flex items-center justify-center w-6 h-6 bg-blue-100 rounded-full -start-3 ring-8 ring-white dark:ring-gray-900 dark:bg-blue-900" {
                    (self.icon)
                };

                div class=(format!("p-4 bg-white border border-gray-200 rounded-lg shadow-xs dark:bg-gray-700 dark:border-gray-600{}", if self.body.is_none() { " items-center justify-between sm:flex" } else { "" })) {
                    @if let Some(body) = &self.body {
                        div class="items-center justify-between mb-3 sm:flex" {
                            time class="mb-1 text-xs font-normal text-gray-400 sm:order-last sm:mb-0" { (self.time) };
                            div class="text-sm font-normal text-gray-500 lex dark:text-gray-300" { (self.title) };
                        };

                        (body)
                    } @else {
                        time class="mb-1 text-xs font-normal text-gray-400 sm:order-last sm:mb-0" { (self.time) };
                        div class="text-sm font-normal text-gray-500 dark:text-gray-300" { (self.title) };
                    }
                };
            };
        }
    }
}

pub struct ActivityLogWidget {
    vanish: bool,
    elements: Vec<ActivityLogElement>,
    after: Option<Box<dyn UIWidget>>,
}

impl ActivityLogWidget {
    pub fn add_element(mut self, element: ActivityLogElement) -> Self {
        self.elements.push(element);
        self
    }

    pub fn add_after<T: UIWidget + 'static>(mut self, after: T) -> Self {
        self.after = Some(Box::new(after));
        self
    }

    pub fn vanish(mut self) -> Self {
        self.vanish = true;
        self
    }
}

impl Render for ActivityLogWidget {
    fn render(&self) -> maud::Markup {
        self.render_with_class("")
    }
}

impl UIWidget for ActivityLogWidget {
    fn can_inherit(&self) -> bool {
        false
    }

    fn base_class(&self) -> Vec<String> {
        Vec::new()
    }

    fn extended_class(&self) -> Vec<String> {
        Vec::new()
    }

    fn render_with_class(&self, _: &str) -> maud::Markup {
        if self.vanish {
            return html! {
                @for element in &self.elements {
                    (element);
                }

                @if let Some(after) = &self.after {
                    (after)
                }
            };
        }

        html! {
            ol class="ml-6 mt-4 relative border-s border-gray-200 dark:border-gray-700" {
                @for element in &self.elements {
                    (element);
                }

                @if let Some(after) = &self.after {
                    (after)
                }
            };
        }
    }
}

#[allow(non_snake_case)]
pub fn ActivityLog() -> ActivityLogWidget {
    ActivityLogWidget {
        vanish: false,
        elements: Vec::new(),
        after: None,
    }
}
