use maud::{Markup, PreEscaped, Render, html};
use prelude::Div;
use primitives::div::DivWidget;

// UI

// Basic Primitives
pub mod color;
pub mod htmx;
pub mod primitives;
pub mod wrapper;

// Complex Components
pub mod components;

// Preludes
pub mod prelude {
    pub use super::UIWidget;
    pub use super::color::*;
    pub use super::primitives::animation::{Animated, Animation, Delay, Duration, Scope, Timing};
    pub use super::primitives::background::{
        Background, BackgroundClip, BackgroundGradient, BackgroundOrigin, BackgroundPosition,
        BackgroundRepeat, BackgroundScrollAttachment, BackgroundSize,
    };
    pub use super::primitives::border::{
        Border, BorderSide, BorderSize, BorderStyle, Outline, OutlineStyle, Ring,
    };
    pub use super::primitives::container::Container;
    pub use super::primitives::cursor::{Action, Cursor, TouchAction};
    pub use super::primitives::display::{
        AlignSelf, Aspect, BoxDecorationBreak, BoxSizing, BreakAfter, BreakBefore, BreakInside,
        BreakInsideValue, BreakValue, Clear, Display, Float, JustifySelf, ObjectFit, Overflow,
        PlaceSelf, ZIndex,
    };
    pub use super::primitives::div::Div;
    pub use super::primitives::filter::{
        BackgroundBlendMode, BlendMode, Blur, Brightness, Contrast, Grayscale, HueRotate, Invert,
        MixBlendMode, Opacity, Saturate, Sepia,
    };
    pub use super::primitives::flex::{
        AlignContent, AlignItems, Column, Direction, DivideStyle, DivideWidth, Flex, FlexBasis,
        FlexGrow, Justify, JustifyItems, Order, Row, Strategy, Wrap,
    };
    pub use super::primitives::grid::{
        Columns, Grid, GridAmount, GridAutoFlow, GridAutoSize, GridElementColumn, GridElementRow,
        GridElementValue,
    };
    pub use super::primitives::height::{Height, MaxHeight, MinHeight};
    pub use super::primitives::image::{Image, Source, Video};
    pub use super::primitives::input::{
        Button, ButtonGroup, Checkbox, CustomRadio, DatePicker, FileInput, Form, FormMethod,
        FormResetButton, FormSubmitButton, HiddenInput, InputAttr, NumberInput, PinInput, Radio,
        Range, Select, TextArea, TextInput, TimePicker, Toggle,
    };
    pub use super::primitives::link::Link;
    pub use super::primitives::list::{HorizontalList, ListEntry, OrderedList, UnorderedList};
    pub use super::primitives::margin::Margin;
    pub use super::primitives::padding::Padding;
    pub use super::primitives::position::{
        ObjectPosition, Position, PositionKind, Resize, Resizeable,
    };
    pub use super::primitives::rounded::Rounded;
    pub use super::primitives::scroll::{Overscroll, Scroll, SnapAlign, SnapType};
    pub use super::primitives::shadow::{DropShadow, Shadow};
    pub use super::primitives::sized::Sized;
    pub use super::primitives::space::{Fraction, ScreenValue, SpaceBetween};
    pub use super::primitives::svg::SVG;
    pub use super::primitives::table::{Caption, Header, Table, TableData, TableHead, TableRow};
    pub use super::primitives::text::{
        AccentColor, Code, DecorationKind, DecorationStyle, DecorationThickness, LetterSpacing,
        LineClamp, LineHeight, ListStyle, NumberStyle, Paragraph, Span, Text, TextAlignment,
        TextContent, TextCursorColor, TextDecoration, TextHyphens, TextOverflow, TextSelection,
        TextTransform, TextWhitespace, TextWordBreak, TextWrap, UnderlineOffset,
        VerticalTextAlignment,
    };
    pub use super::primitives::transform::{
        RenderTransformCPU, RenderTransformGPU, Rotate, Scale, Skew, SkewValue, Transform,
        TransformOrigin,
    };
    pub use super::primitives::visibility::Visibility;
    pub use super::primitives::width::{MaxWidth, MinWidth, Width};
    pub use super::primitives::{
        Context, NoBrowserAppearance, Nothing, Optional, Side, Size, script,
    };
    pub use super::wrapper::{
        _2XLScreen, Hover, LargeScreen, MediumScreen, Screen, SmallScreen, XLScreen,
    };
    pub use maud::Render;
}

/// Generic UI Widget
pub trait UIWidget: Render {
    /// Indicating if the widget supports inheriting classes
    fn can_inherit(&self) -> bool;

    /// Returning the base classes for this widget
    ///
    /// Base here means all classes defining the current widget
    fn base_class(&self) -> Vec<String>;

    /// Return own base classes and all classes below the tree
    fn extended_class(&self) -> Vec<String>;

    /// Render the widget with additional classes
    fn render_with_class(&self, class: &str) -> Markup;
}

/// Implementation for raw HTML with html! macro
impl UIWidget for PreEscaped<String> {
    fn can_inherit(&self) -> bool {
        false
    }

    fn base_class(&self) -> Vec<String> {
        vec![]
    }

    fn extended_class(&self) -> Vec<String> {
        vec![]
    }

    fn render_with_class(&self, _: &str) -> Markup {
        self.render()
    }
}

impl UIWidget for String {
    fn can_inherit(&self) -> bool {
        false
    }

    fn base_class(&self) -> Vec<String> {
        Vec::new()
    }

    fn extended_class(&self) -> Vec<String> {
        Vec::new()
    }

    fn render_with_class(&self, _: &str) -> Markup {
        html!((self))
    }
}

impl UIWidget for &str {
    fn can_inherit(&self) -> bool {
        false
    }

    fn base_class(&self) -> Vec<String> {
        Vec::new()
    }

    fn extended_class(&self) -> Vec<String> {
        Vec::new()
    }

    fn render_with_class(&self, _: &str) -> Markup {
        html!((self))
    }
}

impl<T: UIWidget + 'static> Into<DivWidget> for Vec<T> {
    fn into(self) -> DivWidget {
        let mut div = Div().vanish();

        for e in self {
            div = div.push(e);
        }

        div
    }
}

/// Trait for an element which can add new `attrs`
pub trait AttrExtendable: Sized {
    #[must_use]
    fn add_attr(self, key: &str, val: &str) -> Self;

    /// Set the `id` attribute of an element.
    #[must_use]
    fn id(self, id: &str) -> Self {
        self.add_attr("id", id)
    }
}
